/*
 ***************************************************************************
 * Ralink Tech Inc.
 * 4F, No. 2 Technology 5th Rd.
 * Science-based Industrial Park
 * Hsin-chu, Taiwan, R.O.C.
 *
 * (c) Copyright 2002-2004, Ralink Technology, Inc.
 *
 * All rights reserved. Ralink's source code is an unpublished work and the
 * use of a copyright notice does not imply otherwise. This source code
 * contains confidential trade secret material of Ralink Tech. Any attemp
 * or participation in deciphering, decoding, reverse engineering or in any
 * way altering the source code is stricitly prohibited, unless the prior
 * written consent of Ralink Technology, Inc. is obtained.
 ***************************************************************************

	Module Name:
	rt_chip.c

	Abstract:
	Ralink Wireless driver CHIP related functions

	Revision History:
	Who         When          What
	--------    ----------    ----------------------------------------------
*/


#include "rt_config.h"




/*
========================================================================
Routine Description:
	Initialize specific beacon frame architecture.

Arguments:
	pAd				- WLAN control block pointer

Return Value:
	None

Note:
========================================================================
*/
VOID RtmpChipBcnSpecInit(
	IN RTMP_ADAPTER *pAd)
{
#ifdef MBSS_SUPPORT
#ifdef SPECIFIC_BCN_BUF_SUPPORT
	RTMP_CHIP_OP *pChipOps = &pAd->chipOps;


	pChipOps->FlgIsSupSpecBcnBuf = TRUE;
	pChipOps->BcnMaxHwNum = 16;
	pChipOps->WcidHwRsvNum = 255;

/* 	In 16-MBSS support mode, if AP-Client is enabled, 
	the last 8-MBSS would be occupied for AP-Client using. */
#ifdef APCLI_SUPPORT
	pChipOps->BcnMaxNum = (8 - MAX_MESH_NUM);
#else
	pChipOps->BcnMaxNum = (16 - MAX_MESH_NUM);
#endif // APCLI_SUPPORT //

	pChipOps->BcnMaxHwSize = 0x2000;

	/* It's allowed to use the higher(secordary) 8KB shared memory */
	pChipOps->BcnBase[0] = 0x4000;
	pChipOps->BcnBase[1] = 0x4200;
	pChipOps->BcnBase[2] = 0x4400;
	pChipOps->BcnBase[3] = 0x4600;
	pChipOps->BcnBase[4] = 0x4800;
	pChipOps->BcnBase[5] = 0x4A00;
	pChipOps->BcnBase[6] = 0x4C00;
	pChipOps->BcnBase[7] = 0x4E00;
	pChipOps->BcnBase[8] = 0x5000;
	pChipOps->BcnBase[9] = 0x5200;
	pChipOps->BcnBase[10] = 0x5400;
	pChipOps->BcnBase[11] = 0x5600;
	pChipOps->BcnBase[12] = 0x5800;
	pChipOps->BcnBase[13] = 0x5A00;
	pChipOps->BcnBase[14] = 0x5C00;
	pChipOps->BcnBase[15] = 0x5E00;

	DBGPRINT(RT_DEBUG_TRACE, ("<<<<< Beacon Spec Information: >>>>>\n"));
	DBGPRINT(RT_DEBUG_TRACE, ("\tBcnMaxHwNum = \t%d\n", pChipOps->BcnMaxHwNum));
	DBGPRINT(RT_DEBUG_TRACE, ("\tBcnMaxNum = \t%d\n", pChipOps->BcnMaxNum));
	DBGPRINT(RT_DEBUG_TRACE, ("\tBcnMaxHwSize = \t0x%x\n", pChipOps->BcnMaxHwSize));
	DBGPRINT(RT_DEBUG_TRACE, ("\tWcidHwRsvNum = \t%d\n", pChipOps->WcidHwRsvNum));
	DBGPRINT(RT_DEBUG_TRACE, ("\tBcnBase[0] = \t0x%x\n", pChipOps->BcnBase[0]));
	DBGPRINT(RT_DEBUG_TRACE, ("\tBcnBase[8] = \t0x%x\n", pChipOps->BcnBase[8]));
#endif // SPECIFIC_BCN_BUF_SUPPORT //
#endif // MBSS_SUPPORT //
}


/*
========================================================================
Routine Description:
	Initialize normal beacon frame architecture.

Arguments:
	pAd				- WLAN control block pointer

Return Value:
	None

Note:
========================================================================
*/
VOID RtmpChipBcnInit(
	IN RTMP_ADAPTER *pAd)
{
	RTMP_CHIP_OP *pChipOps = &pAd->chipOps;


	pChipOps->FlgIsSupSpecBcnBuf = FALSE;
	pChipOps->BcnMaxHwNum = 8;
	pChipOps->BcnMaxNum = (pChipOps->BcnMaxHwNum - MAX_MESH_NUM - MAX_APCLI_NUM);
	pChipOps->BcnMaxHwSize = 0x1000;

	pChipOps->BcnBase[0] = 0x7800;
	pChipOps->BcnBase[1] = 0x7A00;
	pChipOps->BcnBase[2] = 0x7C00;
	pChipOps->BcnBase[3] = 0x7E00;
	pChipOps->BcnBase[4] = 0x7200;
	pChipOps->BcnBase[5] = 0x7400;
	pChipOps->BcnBase[6] = 0x5DC0;
	pChipOps->BcnBase[7] = 0x5BC0;

	// If the MAX_MBSSID_NUM is larger than 6, 
	// it shall reserve some WCID space(wcid 222~253) for beacon frames. 
	// -	these wcid 238~253 are reserved for beacon#6(ra6).
	// -	these wcid 222~237 are reserved for beacon#7(ra7).
	if (pChipOps->BcnMaxNum == 8)
		pChipOps->WcidHwRsvNum = 222;
	else if (pChipOps->BcnMaxNum == 7)
		pChipOps->WcidHwRsvNum = 238;
	else
		pChipOps->WcidHwRsvNum = 255;

	DBGPRINT(RT_DEBUG_TRACE, ("<<<<< Beacon Information: >>>>>\n"));
	DBGPRINT(RT_DEBUG_TRACE, ("\tBcnMaxHwNum = \t%d\n", pChipOps->BcnMaxHwNum));
	DBGPRINT(RT_DEBUG_TRACE, ("\tBcnMaxNum = \t%d\n", pChipOps->BcnMaxNum));
	DBGPRINT(RT_DEBUG_TRACE, ("\tBcnMaxHwSize = \t0x%x\n", pChipOps->BcnMaxHwSize));
	DBGPRINT(RT_DEBUG_TRACE, ("\tWcidHwRsvNum = \t%d\n", pChipOps->WcidHwRsvNum));
	DBGPRINT(RT_DEBUG_TRACE, ("\tBcnBase[0] = \t0x%x\n", pChipOps->BcnBase[0]));
	DBGPRINT(RT_DEBUG_TRACE, ("\tBcnBase[1] = \t0x%x\n", pChipOps->BcnBase[1]));
	DBGPRINT(RT_DEBUG_TRACE, ("\tBcnBase[2] = \t0x%x\n", pChipOps->BcnBase[2]));
	DBGPRINT(RT_DEBUG_TRACE, ("\tBcnBase[3] = \t0x%x\n", pChipOps->BcnBase[3]));
	DBGPRINT(RT_DEBUG_TRACE, ("\tBcnBase[4] = \t0x%x\n", pChipOps->BcnBase[4]));
	DBGPRINT(RT_DEBUG_TRACE, ("\tBcnBase[5] = \t0x%x\n", pChipOps->BcnBase[5]));
	DBGPRINT(RT_DEBUG_TRACE, ("\tBcnBase[6] = \t0x%x\n", pChipOps->BcnBase[6]));
	DBGPRINT(RT_DEBUG_TRACE, ("\tBcnBase[7] = \t0x%x\n", pChipOps->BcnBase[7]));
}


/*
========================================================================
Routine Description:
	Initialize chip related information.

Arguments:
	pCB				- WLAN control block pointer

Return Value:
	None

Note:
========================================================================
*/
VOID RtmpChipOpsHook(
	IN VOID			*pCB)
{
	RTMP_ADAPTER *pAd = (RTMP_ADAPTER *)pCB;
	RTMP_CHIP_OP *pChipOps = &pAd->chipOps;

	pChipOps->pRFRegTable = NULL;
	pChipOps->pBBPRegTable = NULL;
	pChipOps->bbpRegTbSize = 0;
	pChipOps->AsicRfInit = NULL;
	pChipOps->AsicBbpInit = NULL;
	pChipOps->AsicRfTurnOn = NULL;
	pChipOps->AsicRfTurnOff = NULL;
	pChipOps->AsicReverseRfFromSleepMode = NULL;
	pChipOps->AsicHaltAction = NULL;
	pChipOps->AsicEeBufferInit = NULL;

	pChipOps->MaxNumOfRfId = 31;
	pChipOps->SnrFormula = SNR_FORMULA1;

#ifdef RTMP_INTERNAL_TX_ALC
	pChipOps->TxAlcTxPowerUpperBound = 45;
#endif // RTMP_INTERNAL_TX_ALC //

	RtmpChipBcnInit(pAd);


	/* We depends on RfICType and MACVersion to assign the corresponding operation callbacks. */
#ifdef RT2880
	if (IS_RT2880(pAd))
	{
		pChipOps->AsicEeBufferInit = RT2880_AsicEeBufferInit;
	}
#endif // RT2880 //


#ifdef RT305x
	pChipOps->AsicEeBufferInit = RT305x_AsicEeBufferInit;

#ifdef RT3352
	//FIXME by Steven: RFIC=RFIC_3022 in some RT3352 board
//	if (pAd->RfIcType == RFIC_3322) {
	if (IS_RT3352(pAd))
	{
#ifdef RTMP_INTERNAL_TX_ALC
		extern TX_POWER_TUNING_ENTRY_STRUCT *TxPowerTuningTable;
		extern TX_POWER_TUNING_ENTRY_STRUCT RT3352_TxPowerTuningTable[];
#endif // RTMP_INTERNAL_TX_ALC //

		pChipOps->MaxNumOfRfId = 63;
		pChipOps->pRFRegTable = RT3352_RFRegTable;
		pChipOps->pBBPRegTable = RT3352_BBPRegTable;
		pChipOps->bbpRegTbSize = RT3352_NUM_BBP_REG_PARMS;
		pChipOps->AsicRfInit = NICInitRT3352RFRegisters;
		pChipOps->AsicBbpInit = NICInitRT3352BbpRegisters;
		pChipOps->AsicMacInit = NICInitRT3352MacRegisters;
		pChipOps->AsicEeBufferInit = RT3352_AsicEeBufferInit;

#ifdef RTMP_INTERNAL_TX_ALC
		pChipOps->TxAlcTxPowerUpperBound = 61;
		TxPowerTuningTable = RT3352_TxPowerTuningTable;
#endif // RTMP_INTERNAL_TX_ALC //

		pChipOps->SnrFormula = SNR_FORMULA2;
		RtmpChipBcnSpecInit(pAd);
	} 
	else
#endif // RT3352 //

	if ((pAd->MACVersion == 0x28720200) && 
		((pAd->RfIcType == RFIC_3320) || (pAd->RfIcType == RFIC_3020) || (pAd->RfIcType == RFIC_3021) || (pAd->RfIcType == RFIC_3022)))
	{
		pChipOps->pRFRegTable = RT305x_RFRegTable;
		pChipOps->pBBPRegTable = RT305x_BBPRegTable;
		pChipOps->bbpRegTbSize = RT305x_NUM_BBP_REG_PARMS;
		pChipOps->AsicRfInit = NICInitRT305xRFRegisters;
	}
#endif // RT305x //




	DBGPRINT(RT_DEBUG_TRACE, ("Chip specific bbpRegTbSize=%d!\n", pChipOps->bbpRegTbSize));
	
}

/* End of rtmp_chip.c */
